/*
 * $Id$
 *
 * ser module that - given sip:email - allows quering LDAPS mail alias
 *
 * Copyright (C) 2004 Swiss Federal Institute of Technology (ETH Zurich)
 * Copyright (C) 2004 Marcel Baur <baur@ethworld.ethz.ch>
 *
 * This module is released under the GNU General Public License
 * Visit http://www.ethworld.ethz.ch for updated and further information.
 */
 
/*
 * This file is part of ser, a free SIP server.
 * Copyright (C) 2001-2003 Fhg Fokus
 *
 * ser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * ser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * History:
 * --------
 * 2004-12-01 Initial version (Marcel Baur)
 */




#include "../../sr_module.h"
#include "../../db/db.h"
#include "../../action.h"
#include <stdio.h>
#include <ldap.h>

#define TABLE_SETUP "ethworld_setup"


static db_con_t *db_con;
static db_func_t *db_f;



MODULE_VERSION static int ldap_alias_f (struct sip_msg *, char *, char *);

static int mod_init (void);

/* module parameters */

static char *LDAPS_User = "username";
static char *LDAPS_Password = "password";
static char *LDAPS_Bind_dn = "cn=username,OU=org,dc=d,dc=my,dc=org";
static char *LDAPS_Server = "ldaps.my.org";
static char *LDAPS_Base_dn = "dc=d,dc=my,dc=org";
static char *LDAPS_Attribute = "sAMAccountName";
static char *LDAPS_FilterPrefix = "proxyAddresses=smtp:";
static char *APPEND_Domain = "my.org";
static int LDAPS_Port = 636;
static int LDAPS_ProtocolVersion = LDAP_VERSION3;

static char *FIELD_INETPHONENO = "inetPhoneNo";
static char *FIELD_USERNAME = "username";

static char *db_url = DEFAULT_DB_URL;


static cmd_export_t cmds[] = {
  {"ldaps_lookup", ldap_alias_f, 0, 0, REQUEST_ROUTE},
  {0, 0, 0, 0, 0}
};

static param_export_t params[] = {
  {"ldaps_server", STR_PARAM, &LDAPS_Server},
  {"ldaps_port", INT_PARAM, &LDAPS_Port},
  {"ldaps_base", STR_PARAM, &LDAPS_Base_dn},
  {"ldaps_attribute", STR_PARAM, &LDAPS_Attribute},
  {"append_domain", STR_PARAM, &APPEND_Domain},
  {"ldaps_user", STR_PARAM, &LDAPS_User},
  {"ldaps_password", STR_PARAM, &LDAPS_Password},
  {"ldaps_bind_dn", STR_PARAM, &LDAPS_Bind_dn},
  {"ldaps_filterprefix", STR_PARAM, &LDAPS_FilterPrefix},
  {"field_inetPhoneNo", STR_PARAM, &FIELD_INETPHONENO},
  {"field_username", STR_PARAM, &FIELD_USERNAME},
  {0, 0, 0}
};

struct module_exports exports = {
  "ldap_alias",
  cmds,
  params,

  mod_init,			/* module initialization function */
  0,				/* response function */
  0,				/* destroy function */
  0,				/* oncancel function */
  0				/* per-child init function */
};


#define RESULT_YES 1
#define RESULT_DROP_PACKAGE 0
#define RESULT_NO -1

static int
mod_init (void)
{
  int res, retval;
  LOG (L_ERR, "LDAPS Module\n"
       "Copyright (C) 2004 Marcel Baur <baur@ethworld.ethz.ch>\n"
       "Copyright (C) 2004 Swiss Federal Institute of Technology (ETH Zurich)\n");

  res = bind_dbmod (db_url, db_f);
  if (res < 0)
    {
      LOG (L_ERR, "ldaps: Unable to bind to db module (err=%i)\n", res);
      return -1;
    }
  db_con = db_f->init (db_url);
  if (db_con == NULL)
    {
      LOG (L_ERR, "ldaps: Error in db_init");
      retval = -1;
    }
  else
    {
      retval = 0;
    }
  return retval;
}


/* ------------------------------------------------------------------- */
/* given a header field 'To: <sip:user@host>',
 * alias email user@host in LDAPS and
 * return sip url 'sip:32697@my.org'
 */

static char *
get_filter (str to)
{
  char tmp[255];
  int start, stop, i;

  start = 0;
  i = 0;

  if (strncmp (to.s, "sip:", 4) == 0)
    {
      start = 0;
    }
  else
    {
      while (to.s[i] != '<')
	{
	  start++;
	  i++;
	}
      start++;
    }

  while ((to.s[i] != '>') & (i < to.len))
    {
      i++;
    }
  stop = i;

  if (strncmp (&to.s[start], "sip:", 4) == 0)
    {
      start += 4;
    }

  if (stop < start)
    {
      return NULL;
    }

  memset (tmp, 0, sizeof (tmp));
  strcpy (tmp, LDAPS_FilterPrefix);
  strncat (tmp, &to.s[start], stop - start);
  LOG (L_ERR, "get_filter is '%s'", tmp);
  return strdup (tmp);
}

/* ------------------------------------------------------------------ */
/* set uri                                                            */

static void
set_uri (const char *uri, struct sip_msg *msg)
{
  struct action act;
  int ret;

  memset (&act, 0, sizeof (act));
  act.type = SET_URI_T;
  act.p1_type = STRING_ST;
  act.p1.string = strdup (uri);
  ret = do_action (&act, msg);

  if (ret < 0)
    {
      LOG (L_ERR, "ERROR: unable to set uri (error=%i)\n", ret);
    }
}

/* ------------------------------------------------------------------ */
/* we registered only 1 param, so we ignore str2 */

static char *
username_to_inet (const char *username)
{
  int result, retval;
  db_key_t _k[2];
  db_val_t _v[2];
  db_key_t _c[2];
  db_res_t *res;
  int n, nc, anz;
  char tmp[255];

  db_val_t *values;
  db_row_t *rows;
  int i;

  retval = RESULT_NO;

  /* SELECT */
  _c[0] = FIELD_INETPHONENO;
  nc = 1;

  /* FROM */
  db_f->use_table (db_con, TABLE_SETUP);

  /* WHERE */
  _k[0] = FIELD_USERNAME;
  VAL_TYPE (_v) = DB_STR;
  VAL_NULL (_v) = 0;
  VAL_STR (_v).s = (char *) username;
  VAL_STR (_v).len = strlen (username);
  n = 1;

  /* execute query */
  result = db_f->query (db_con, _k, 0, _v, _c, n, nc, 0, &res);
  if (result < 0)
    {
      LOG (L_ERR, "username_to_inet: error %i querying db\n", result);
      return NULL;
    }

  anz = RES_ROW_N (res);

  if (anz == 0)
    {
      /* no entry found */
      db_f->free_result (db_con, res);
      LOG (L_ERR, "No %s entry for '%s'", FIELD_INETPHONENO, username);
      return NULL;
    }
  else
    {
      /* get result */
      rows = RES_ROWS (res);
      values = ROW_VALUES (rows);

      for (i = 0; i < anz; i++)
	{
	  strncpy (tmp, VAL_STR (values).s, sizeof (tmp));
	}
    }

  db_f->free_result (db_con, res);
  return strdup (tmp);
}

/* ------------------------------------------------------------------ */
/* ldap_alias                                                         */

static char *
ldaps_lookup (char *filter)
{
  int rc, x, i;
  LDAP *ld;
  LDAPMessage *res, *ent;
  char *a;
  char *attrs[] = { LDAPS_Attribute, NULL };
  BerElement *ber;

  char **val;
  char tmp[255];

  ld = ldap_init (LDAPS_Server, LDAPS_Port);
  if (ld == NULL)
    {
      LOG (L_ERR, "ldaps: ldap_init failed");
      return NULL;
    }

  /* must be version 3 */
  rc =
    ldap_set_option (ld, LDAP_OPT_PROTOCOL_VERSION, &LDAPS_ProtocolVersion);
  if (rc != LDAP_SUCCESS)
    {
      LOG (L_ERR, "ldaps: setting LDAP_OPT_PROTOCOL_VERSION failed (rc=%i)",
	   rc);
      LOG (L_ERR, "ldaps: %s", ldap_err2string (rc));
      return NULL;
    }

  /* set TLS option */
  x = LDAP_OPT_X_TLS_HARD;
  rc = ldap_set_option (ld, LDAP_OPT_X_TLS, &x);
  if (rc != LDAP_SUCCESS)
    {
      LOG (L_ERR, "ldaps: setting LDAP_OPT_X_TLS failed (rc=%i)", rc);
      LOG (L_ERR, "ldaps: %s", ldap_err2string (rc));
      return NULL;
    }

  rc = ldap_bind_s (ld, LDAPS_Bind_dn, LDAPS_Password, LDAP_AUTH_SIMPLE);
  if (rc != LDAP_SUCCESS)
    {
      LOG (L_ERR, "ldaps: ldap_bind failed (rc=%i)", rc);
      LOG (L_ERR, "ldaps: %s", ldap_err2string (rc));
      return NULL;
    }

  rc =
    ldap_search_s (ld, LDAPS_Base_dn, LDAP_SCOPE_SUBTREE, filter, attrs, 0,
		   &res);
  if (rc != LDAP_SUCCESS)
    {
      LOG (L_ERR, "ldaps: ldap_search_s failed (rc=%i)", rc);
      LOG (L_ERR, "ldaps: %s", ldap_err2string (rc));
      return NULL;
    }

  ent = ldap_first_entry (ld, res);
  if (ent == NULL)
    {
      LOG (L_ERR, "ldaps: No entries found");
      return NULL;
    }

  a = ldap_first_attribute (ld, ent, &ber);
  if (a == NULL)
    {
      LOG (L_ERR, "ldaps: attribute is NULL");
      return NULL;
    }

  val = ldap_get_values (ld, ent, a);
  if (val == NULL)
    {
      LOG (L_ERR, "ldaps: values is NULL");
      return NULL;
    }

  for (i = 0; val[i] != NULL; i++)
    {
      strncpy (tmp, val[0], sizeof (tmp));
      LOG (L_ERR, "ldaps: got '%s'", val[i]);
    }

  ldap_value_free (val);

  return strdup (tmp);
}

/* ------------------------------------------------------------------ */

static int
ldap_alias_f (struct sip_msg *mmm, char *m1, char *m2)
{
  struct to_body *receiver;
  char *filter, *username, *inet;
  char uri[255];

  receiver = get_to (mmm);

  filter = get_filter (receiver->uri);
  if (filter == NULL)
    {
      LOG (L_ERR, "ldaps: filter is NULL");
      return RESULT_NO;
    }

  username = ldaps_lookup (filter);
  if (username == NULL)
    {
      LOG (L_ERR, "ldaps: username is NULL");
      return RESULT_NO;
    }

  inet = username_to_inet (username);
  if (inet == NULL)
    {
      LOG (L_ERR, "ldaps: no inet for username '%s'", username);
      return RESULT_NO;
    }

  strcpy (uri, "sip:");
  strcat (uri, inet);
  strcat (uri, "@");
  strcat (uri, APPEND_Domain);

  set_uri (uri, mmm);

  LOG (L_ERR, "ldaps_alias: %s", uri);
  return RESULT_YES;
}

/* ------------------------------------------------------------------ */
